"""
"Lookup" cog for discordbot - https://github.com/pvyParts/allianceauth-discordbot
"""

# Standard Library
import csv
import io
import logging

# Third Party
from discord import File
from discord.colour import Color
from discord.commands import SlashCommandGroup, option
from discord.embeds import Embed
from discord.ext import commands

# Django
from django.conf import settings
from django.contrib.auth.models import User
from django.core.exceptions import ObjectDoesNotExist

# Alliance Auth
from allianceauth.eveonline.evelinks import evewho
from allianceauth.eveonline.models import EveCharacter, EveCorporationInfo

# Alliance Auth Discord Bot
from aadiscordbot import app_settings
from aadiscordbot.app_settings import aastatistics_active
from aadiscordbot.cogs.utils.autocompletes import (
    search_characters,
    search_corporations_on_characters,
)
from aadiscordbot.cogs.utils.decorators import (
    is_guild_managed,
    message_in_channels,
    sender_has_any_perm,
)

# Terra Nanotech Discordbot Cogs
from tnnt_discordbot_cogs.helper import unload_cog
from tnnt_discordbot_cogs.models.setting import Setting

logger = logging.getLogger(__name__)


class Lookup(commands.Cog):
    """
    All about users!
    """

    def __init__(self, bot):
        """
        Initialize the Lookup cog.

        :param bot:
        :type bot:
        """

        self.bot = bot

    lookup_commands = SlashCommandGroup(
        "lookup", "Server Admin Commands", guild_ids=app_settings.get_all_servers()
    )

    @staticmethod
    def _get_lookup_channels() -> list:
        """
        Get the lookup channels from the settings.

        :return: List of lookup channels or an empty list if none are set.
        :rtype: dict
        """

        lookup_channels = Setting.get_setting(Setting.Field.LOOKUP_CHANNELS.value).all()

        return (
            [channel.channel for channel in lookup_channels if channel is not None]
            if lookup_channels
            else []
        )

    @staticmethod
    def get_csv(input_name):
        """
        Generates a CSV file of all known alts for a given character name.

        :param input_name:
        :type input_name:
        :return:
        :rtype:
        """

        char = EveCharacter.objects.get(character_name=input_name)
        alts = char.character_ownership.user.character_ownerships.all().select_related(
            "character"
        )
        buffer = io.StringIO()
        writer = csv.writer(buffer)
        header = [
            "Character ID",
            "Character Name",
            "Corporation ID",
            "Corporation Name",
            "Alliance ID",
            "Alliance Name",
        ]
        writer.writerow(header)

        for a in alts:
            writer.writerow(
                [
                    a.character.character_id,
                    a.character.character_name,
                    a.character.corporation_id,
                    a.character.corporation_name,
                    a.character.alliance_id,
                    a.character.alliance_name,
                ]
            )

        buffer.seek(0)

        return File(buffer, f"{input_name}_known_alts.csv")

    @staticmethod
    def get_lookup_embed(input_name):
        """
        Generates an embed with information about a character.

        :param input_name:
        :type input_name:
        :return:
        :rtype:
        """

        embed = Embed(title=f"Character Lookup: {input_name}")

        try:
            char = EveCharacter.objects.get(character_name=input_name)

            try:
                main = char.character_ownership.user.profile.main_character
                state = char.character_ownership.user.profile.state.name
                groups = char.character_ownership.user.groups.all().values_list(
                    "name", flat=True
                )

                try:
                    discord_string = f"<@{char.character_ownership.user.discord.uid}>"
                except Exception as e:
                    logger.error(e)

                    discord_string = "unknown"

                if aastatistics_active():
                    alts = (
                        char.character_ownership.user.character_ownerships.all()
                        .select_related("character", "character_stats")
                        .values_list(
                            "character__character_name",
                            "character__corporation_ticker",
                            "character__character_id",
                            "character__corporation_id",
                            "character__character_stats__zk_12m",
                            "character__character_stats__zk_3m",
                        )
                    )
                    zk12 = 0
                    zk3 = 0
                else:
                    alts = (
                        char.character_ownership.user.character_ownerships.all()
                        .select_related("character")
                        .values_list(
                            "character__character_name",
                            "character__corporation_ticker",
                            "character__character_id",
                            "character__corporation_id",
                        )
                    )
                    zk12 = "Not Installed"
                    zk3 = "Not Installed"

                if aastatistics_active():
                    for alt in alts:
                        if alt[4]:
                            zk12 += alt[4]
                            zk3 += alt[5]

                embed.colour = Color.blue()
                embed.description = f"**{char}** is linked to **{main} [{main.corporation_ticker}]** (State: {state})"

                alt_list = [
                    f"[{a[0]}]({evewho.character_url(a[2])}) [[{a[1]}]({evewho.corporation_url(a[3])})]"
                    for a in alts
                ]

                for idx, names in enumerate(
                    [alt_list[i : i + 6] for i in range(0, len(alt_list), 6)]
                ):
                    if idx < 6:
                        embed.add_field(
                            name=f"Linked Characters {idx + 1}",
                            value="\n".join(names),
                            inline=False,
                        )
                    else:
                        embed.add_field(
                            name=f"Linked Characters {idx} **(Discord Limited There are More)**",
                            value="\n".join(names),
                            inline=False,
                        )
                        break

                if len(groups) > 0:
                    embed.add_field(
                        name="Groups", value="\n".join(groups), inline=False
                    )

                if aastatistics_active():
                    embed.add_field(name="12m Kills", value=zk12, inline=True)
                    embed.add_field(name="3m Kills", value=zk3, inline=True)

                embed.add_field(name="Discord Link", value=discord_string, inline=False)

                return embed
            except ObjectDoesNotExist:
                users = char.ownership_records.values("user")
                users = User.objects.filter(id__in=users)
                characters = EveCharacter.objects.filter(
                    ownership_records__user__in=users
                ).distinct()
                embed = Embed(title="Character Lookup")
                embed.colour = Color.blue()

                embed.description = f"**{char}** is unlinked. Searching for any characters linked to known users"
                user_names = [f"{user.username}" for user in users]

                if len(user_names) == 0:
                    user_names = "No user links found"
                else:
                    user_names = ", ".join(user_names)

                embed.add_field(name="Old Users", value=user_names, inline=False)

                alt_list = [
                    f"[{a.character_name}]({evewho.character_url(a.character_id)}) *[[{a.corporation_ticker}]({evewho.corporation_url(a.corporation_id)})]*"
                    for a in characters
                ]

                for idx, names in enumerate(
                    [alt_list[i : i + 6] for i in range(0, len(alt_list), 6)]
                ):
                    if idx < 6:
                        embed.add_field(
                            name=f"Found Characters {idx + 1}",
                            value=", ".join(names),
                            inline=False,
                        )
                    else:
                        embed.add_field(
                            name=f"Found Characters {idx} **( Discord Limited There are More )**",
                            value=", ".join(names),
                            inline=False,
                        )
                        break

                return embed

        except EveCharacter.DoesNotExist:
            embed.colour = Color.red()

            embed.description = (
                f"Character **{input_name}** does not exist in our Auth system"
            )

            return embed

    @staticmethod
    def build_altcorp_embeds(input_name):
        chars = EveCharacter.objects.filter(corporation_name=input_name)

        if chars.count():
            corp_id = 0
            own_ids = [settings.DISCORD_BOT_MEMBER_ALLIANCES]
            alts_in_corp = []
            knowns = 0

            for c in chars:
                corp_id = c.corporation_id

                if c.alliance_id not in own_ids:
                    alts_in_corp.append(c)

            mains = {}

            for a in alts_in_corp:
                try:
                    main = a.character_ownership.user.profile.main_character

                    if main.character_id not in mains:
                        mains[main.character_id] = [main, 0]

                    mains[main.character_id][1] += 1
                    knowns += 1
                except Exception:
                    # logger.error(e)
                    pass

            output = []
            base_string = "[{}]({}) [[{}]({})] has {} character{}"

            for k, m in mains.items():
                output.append(
                    base_string.format(
                        m[0],
                        evewho.character_url(m[0].character_id),
                        m[0].corporation_ticker,
                        evewho.corporation_url(m[0].corporation_id),
                        m[1],
                        "s" if m[1] > 1 else "",
                    )
                )
            embeds = []

            corp_info = EveCorporationInfo.provider.get_corporation(corp_id)

            msg = (
                f"**[[{corp_info.ticker}]({evewho.corporation_url(corp_id)})]** has {corp_info.members} members:\n\n"
                "```diff\n"
                f"+Known Members     : {knowns}\n"
                f"-Unknowns          : {corp_info.members - knowns}```"
            )

            _header = Embed(title=input_name, description=msg)

            embeds.append(_header)

            for strings in [output[i : i + 10] for i in range(0, len(output), 10)]:
                embed = Embed(title=input_name)
                embed.colour = Color.blue()
                embed.description = "\n".join(strings)
                embeds.append(embed)

            return embeds

    @lookup_commands.command(
        name="character",
        description="Looks up a character in the Auth system and returns information about them.",
        guild_ids=app_settings.get_all_servers(),
    )
    @is_guild_managed()
    @sender_has_any_perm(
        [
            "corputils.view_alliance_corpstats",
            "corpstats.view_alliance_corpstats",
            "aadiscordbot.member_command_access",
        ]
    )
    # @message_in_channels(settings.ADMIN_DISCORD_BOT_CHANNELS)
    @message_in_channels(channels=_get_lookup_channels())
    @option(
        "character",
        description="Search for a character",
        autocomplete=search_characters,
    )
    @option("gib_csv", description="Output a CSV of all characters")
    async def slash_lookup(self, ctx, character: str, gib_csv: bool = False):
        """
        Looks up a character in the Auth system and returns information about them.
        Input: a Eve Character Name
        Example: `/lookup character: John Doe gib_csv: True`

        :param ctx:
        :type ctx:
        :param character:
        :type character:
        :param gib_csv:
        :type gib_csv:
        :return:
        :rtype:
        """

        await ctx.defer()

        if gib_csv:
            file = self.get_csv(character)

            return await ctx.respond(embed=self.get_lookup_embed(character), file=file)

        return await ctx.respond(embed=self.get_lookup_embed(character))

    @lookup_commands.command(
        name="corporation",
        description="Looks up a corporation and returns information about its members.",
        guild_ids=app_settings.get_all_servers(),
    )
    @is_guild_managed()
    @sender_has_any_perm(["aadiscordbot.member_command_access"])
    # @message_in_channels(settings.ADMIN_DISCORD_BOT_CHANNELS)
    @message_in_channels(channels=_get_lookup_channels())
    @option(
        "corporation",
        description="Search for a corporation",
        autocomplete=search_corporations_on_characters,
    )
    async def slash_altcorp(self, ctx, corporation: str):
        """
        Gets Auth data about an altcorp

        :param ctx:
        :type ctx:
        :param corporation:
        :type corporation:
        :return:
        :rtype:
        """

        await ctx.defer()

        embeds = self.build_altcorp_embeds(corporation)

        if len(embeds):
            e = embeds.pop(0)

            await ctx.respond(embed=e)

            for e in embeds:
                await ctx.send(embed=e)
        else:
            await ctx.respond("No Members Found!")


def setup(bot):
    """
    Set up the cog

    :param bot:
    :type bot:
    :return:
    :rtype:
    """

    # Unload any other Lookup cog
    unload_cog(bot=bot, cog_name="Lookup")

    # Unload the Members cog from `aadiscordbot`
    unload_cog(bot=bot, cog_name="Members")

    # Add the Lookup cog to the bot
    bot.add_cog(Lookup(bot))



"""
Settings model for the TNNT Discord bot.
"""

# Standard Library
from typing import Any

# Third Party
from solo.models import SingletonModel

# Django
from django.db import models
from django.utils.translation import gettext_lazy as _

# Alliance Auth Discord Bot
from aadiscordbot.models import Channels


class Setting(SingletonModel):
    """
    Settings model for the TNNT Discord bot.
    This model is used to store settings that are not expected to change frequently.
    """

    class Field(models.TextChoices):
        """
        Choices for the setting field.
        """

        APPLICANT_ROLE_NAME = "applicant_role_name", _("Applicant Role Name")
        APPLICANT_ROLE_ID = "applicant_role_id", _("Applicant Role ID")
        RECRUITING_CHANNEL = "recruiting_channel", _("Recruiting Channel")
        RECRUITER_ROLE_ID = "recruiter_role_id", _("Recruiter Role ID")
        LEADERSHIP_ROLE_ID = "leadership_role_id", _("Leadership Role ID")
        WELCOME_CHANNEL_AUTHENTICATED = "welcome_channel_authenticated", _(
            "Welcome Channel (Authenticated User)"
        )
        WELCOME_CHANNEL_UNAUTHENTICATED = "welcome_channel_unauthenticated", _(
            "Welcome Channel (Unauthenticated User)"
        )
        WELCOME_ROLES_EXCLUDED = "welcome_roles_excluded", _("Roles Excluded")
        LOOKUP_CHANNELS = "lookup_channels", _("Lookup Channels")

    # Recruitment Cog Settings
    applicant_role_name = models.CharField(
        max_length=100,
        default="",
        verbose_name=Field.APPLICANT_ROLE_NAME.label,  # pylint: disable=no-member,
        help_text=_("The name of the role for applicants in the Discord server."),
    )

    applicant_role_id = models.PositiveBigIntegerField(
        null=True,
        default=None,
        verbose_name=Field.APPLICANT_ROLE_ID.label,  # pylint: disable=no-member
        help_text=_("The ID of the role for applicants in the Discord server."),
    )

    recruiting_channel = models.ForeignKey(
        to=Channels,
        related_name="+",
        null=True,
        on_delete=models.SET_NULL,
        verbose_name=Field.RECRUITING_CHANNEL.label,  # pylint: disable=no-member
        help_text=_("The channel used for recruiting in the Discord server."),
    )

    recruiter_role_id = models.PositiveBigIntegerField(
        null=True,
        default=None,
        verbose_name=Field.RECRUITER_ROLE_ID.label,  # pylint: disable=no-member
        help_text=_("The ID of the role for recruiters in the Discord server."),
    )

    leadership_role_id = models.PositiveBigIntegerField(
        null=True,
        default=None,
        verbose_name=Field.LEADERSHIP_ROLE_ID.label,  # pylint: disable=no-member
        help_text=_("The ID of the role for leadership in the Discord server."),
    )

    # Welcome Cog Settings
    welcome_channel_authenticated = models.ForeignKey(
        to=Channels,
        related_name="+",
        null=True,
        on_delete=models.SET_NULL,
        verbose_name=Field.WELCOME_CHANNEL_AUTHENTICATED.label,  # pylint: disable=no-member
        help_text=_(
            "The channel to send welcome messages for authenticated users in the Discord server."
        ),
    )

    welcome_channel_unauthenticated = models.ForeignKey(
        to=Channels,
        related_name="+",
        null=True,
        on_delete=models.SET_NULL,
        verbose_name=Field.WELCOME_CHANNEL_UNAUTHENTICATED.label,  # pylint: disable=no-member
        help_text=_(
            "The channel to send welcome messages for unauthenticated users in the Discord server."
        ),
    )

    welcome_roles_excluded = models.CharField(
        max_length=255,
        default="Member",
        verbose_name=Field.WELCOME_ROLES_EXCLUDED.label,  # pylint: disable=no-member
        help_text=_(
            "Comma-separated list of Discord roles that are excluded from the welcome message."
        ),
    )

    lookup_channels = models.ManyToManyField(
        to=Channels,
        related_name="lookup_channels",
        blank=True,
        verbose_name=Field.LOOKUP_CHANNELS.label,  # pylint: disable=no-member
        help_text=_("Channels in which the `/lookup` command can be used."),
    )

    class Meta:
        """
        Meta options for the Setting model.
        """

        default_permissions = ()
        verbose_name = _("Setting")
        verbose_name_plural = _("Settings")

    def __str__(self):
        """
        String representation of the Setting model.

        :return:
        :rtype:
        """

        return str(_("Settings"))

    @staticmethod
    def get_setting(setting_key: str) -> Any:
        """
        Get the setting value for a given setting key

        :param setting_key: The setting key
        :type setting_key: str
        :return: The setting value
        :rtype: Any
        """

        try:
            return getattr(Setting.get_solo(), setting_key)
        except AttributeError as exc:
            raise KeyError(f"Setting key '{setting_key}' does not exist.") from exc



"""
"Auth" cog for discordbot - https://github.com/pvyParts/allianceauth-discordbot
"""

# Standard Library
import logging

# Third Party
from discord.colour import Color
from discord.embeds import Embed
from discord.ext import commands

# Django
from django.conf import settings

# Alliance Auth
from allianceauth.eveonline.evelinks.eveimageserver import (
    alliance_logo_url,
    corporation_logo_url,
)

# Alliance Auth Discord Bot
from aadiscordbot.app_settings import get_site_url

# Terra Nanotech Templates
from tnnt_templates.app_settings import AppSettings

# Terra Nanotech Discordbot Cogs
from tnnt_discordbot_cogs.helper import unload_cog

logger = logging.getLogger(__name__)


class Auth(commands.Cog):
    """
    A Collection of Authentication Tools for Alliance Auth
    """

    def __init__(self, bot):
        """
        Initialize the Auth cog.

        :param bot:
        :type bot:
        """

        self.bot = bot

    @commands.slash_command(
        name="auth",
        description="Returns a link to TN-NT Auth",
        guild_ids=[int(settings.DISCORD_GUILD_ID)],
    )
    async def auth(self, ctx):
        """
        Returns a link to TN-NT Auth
        Used by many other Bots and is a common command that users will attempt to run.

        :param ctx:
        :type ctx:
        :return:
        :rtype:
        """

        await ctx.defer()

        auth_url = get_site_url()
        embed = Embed(title="Terra Nanotech Auth")

        try:
            if settings.TNNT_TEMPLATE_ENTITY_ID == 1:
                aa_icon = f"{auth_url}/static/allianceauth/icons/allianceauth.png"

                embed.set_thumbnail(url=aa_icon)
            else:
                entity_type = AppSettings.TNNT_TEMPLATE_ENTITY_TYPE
                entity_id = AppSettings.TNNT_TEMPLATE_ENTITY_ID

                logo_funcs = {
                    "alliance": (alliance_logo_url, "alliance_id"),
                    "corporation": (corporation_logo_url, "corporation_id"),
                }

                if entity_type in logo_funcs:
                    func, arg_name = logo_funcs[entity_type]

                    embed.set_thumbnail(url=func(**{arg_name: entity_id}, size=256))
        except AttributeError:
            pass

        embed.colour = Color.blue()

        embed.description = (
            "All authentication functions for this Discord "
            "server are handled through our Auth system."
        )

        embed.add_field(name="Auth Link", value=auth_url, inline=False)

        return await ctx.respond(embed=embed)


def setup(bot):
    """
    Set up the cog

    :param bot:
    :type bot:
    :return:
    :rtype:
    """

    # Unload the Auth cog from `aadiscordbot`, so we can load our own.
    unload_cog(bot=bot, cog_name="Auth")

    # Add the Auth cog to the bot
    bot.add_cog(Auth(bot))
